#ifndef APPROX_H
#define APPROX_H

#include <config.h>

// Include some standard libraries
#include <cmath>
#include <iostream>
#include <fstream>
#include <string>
#include <stdlib.h>


// Include some seq++ libraries
#include <seqpp/SequenceSet.h>
#include <seqpp/PhasedMarkov.h>
#include <seqpp/Markov.h>
#include <seqpp/Sequence.h>
#include <seqpp/Translator.h>

// For The GSL, previously installed
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_sort_vector.h>

// Different usefull tools
#include "libext/Eispack.h"
#include "Tools.h"

using namespace std;

// Commentary in the shell
extern int VERBOSE;

//===========================================================================================
//===========================================================================================

class approx
{

//===========================================================================================

 private:

  int _order;             //Order of the Markov Model
  Translator * _alphabet; //Alphabet used (example : {a, c, g, t})
  short _sa;              //Size of the alphabet
  double* _PI;            //Transition matrix
  int _sPI;               //Size of the transition matrix (=size of the alphabet power (order+1))  
  double* _mu;            //_mu[i]=(1/_sa) for all i
  double* _MU;            //Stationary distribution (for words of _order letters)
  double _K;              //Constant K (psi-mixing)(rate of ergodicity)
  double _nu;             //Constant nu (psi-mixing)(rate of ergodicity)
  double _length;         //Length of the sequence
 
//===========================================================================================
  
 public: 

  //--------------//
  // Constructors //
  //--------------//

  approx(short order, double length, char* fm, char* fa, ofstream& fr, const gsl_rng * r);
  //Compute the transition matrix given in the file fm
  //This is an example of file fm : (the two last lines are not necessary)
  //=======================================
  //# 1 <- Order of the phased Markov chain
  //# 1 <- Phase
  //# 4 <- Alphabet size
  //# Transition matrix:
  //0.2 0.3 0.2 0.3
  //0.1 0.2 0.3 0.4
  //0.3 0.3 0.2 0.2
  //0.2 0.2 0.3 0.3
  //# Stationnary Probability:
  //0.2008929 0.2455357 0.2544643 0.2991071
  //=======================================
  //Compute the second eigenvalue and the stationary distribution
  //order is the order of the Markov model
  //length is the sequence length
  //File fa contains the alphabet used
  //File fr contains the results
  //r is gsl parameter
  
  
  approx(short order, ofstream& fr, char*fs, char* fa, const gsl_rng * r);
  //Compute the transition matrix estimated on the sequence given in file fs
  //Compute the second eigenvalue and the stationary distribution
  //order is the order of the Markov model
  //File fr contains the results
  //File fa contains the alphabet used
  

  //------------//
  // Destructor //
  //------------//

  ~approx();                                  
  
  //----------//
  // Methods //
  //----------//

  void printPI ();
  //Print the transition matrix

  void printPI (ofstream& fm);
  //Print the transition matrix in the file fm

  double tell_length ();
  //Return the sequence length
  
  double phi(double l);                 
  //Function phi (or psi) : K*(nu power l)
  
  double probawll(int w, char* A, int n);        
  //Probability of the w last letters of the word A of size n
  
  double epsilon_one(char* A, int n);             
  //Return epsilon 1 for a phi-mixing
  //A is a word of size n
  
  double compute_la(double p, int n);
  //Compute lA
  //p is the probability of a word of size n
  
  double epsilon_two(double p, int n);
  //Return epsilon 2 with p probability of the word and n length of the word
  //p is the probability of a word of size n
  
  double varepsilon(double p, int n);
  //Return inf (Dp(A)+phi(D))
  //p is the probability of a word of size n
  
  double proposition_phi(double p, int n, double t, char* A);
  //Return proposition ?
  //p is the probability of the word A of size n
  //t is the sequence length
  
  double proposition_psi(double p, int n, double t, char* A);
  //Return min { proposition ?, Abadi thesis }
  //p is the probability of the word A of size n
  //t is the sequence length
 
  double psiepsilon_one(char* A, int n);             
  //Return epsilon 1 for a psi-mixing for a word A of size n
  
  Translator * give_alpha() const;
  //Pour pouvoir se servir du membre priv _alphabet
  
  int give_word(char * A, char* wf,int j);
  //Give the jth word (length lj) of file wf
  //A contains this word
 
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // PHI mlanges   
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  
  //In all the following methods, we have :
  //s1  = (k-1)!
  //lx  = log(t(p+phi(la))
  //le  = log(epsilon)
  //la  = la of epsilon 2
  //p1  = error in 0 (proposition phi)
  //n   = length of the word
  //p   = probability of the word
  //par = parameter of the Poisson law

  double error_k 
  (int k,double s1,double lx, double le, double la, double p1, int n, double p, double par);
  //Compute and return error in k
  
  int print_error_k(char* wf,int k,ofstream& fr);
  //Print error in k in a file fr
  //wf is the file of words 
  
  double sum_errors_k
  (int k,double s1,double lx, double le, double la, double p1, int n, double p, double par);
  //Compute and return the sum of errors for errors from k to 2t/n

  int print_sum_errors_k(char* wf,int k,ofstream& fr);
  //Print the sum of errors for errors from k to 2t/n in file fr
  
  int print_u(char* wf, double ds,ofstream& fr, ofstream& fu);
  //Compute the u and print the sum of errors for errors from u to 2t/n
  //where u is such that P(N>u)<ds (and P(N>u-1)>ds)
  //ds = degree of significativity
  //File fr contains the results
  //File fu contains the u 

  double k_sum_errors
  (int k,double lx, double le, double la, double p1, int n, double p, double par);
  //Compute and return the sum of errors for errors from 0 to k

  int print_k_sum_errors(char* wf,int k,ofstream& fr);
  //Print the sum of errors for errors from 0 to k in file fr
    
  int u_print(char* wf, double ds,ofstream& fr, ofstream& fu);
  //Compute the u and print the sum of errors for errors from 0 to u
  //where u is such that P(N<u)<ds (and P(N>u+1)>ds)
  //ds = degree of significativity
  //File fr contains the results
  //File fu contains the u 

  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // PSI mlanges  
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  
  //In all the following methods, we have :
  //s1  = (k-1)!
  //lx  = log(t(p+phi(la))
  //le  = log(epsilon)
  //p3  = error in 0 (proposition psi)
  //n   = length of the word
  //p   = probability of the word
  //par = parameter of the Poisson law
  
  double psi_error_k 
  (int k,double s1,double lx, double le, double p3, int n, double p, double par);
  //Compute and return error in k
 
  int psi_print_error_k(char* wf,int k,ofstream& fr);
  //Print error in a file fr
  //wf is the file of words

  double psi_sum_errors_k
  (int k,double s1,double lx, double le, double p3, int n, double p, double par);
  //Compute and return the sum of errors for errors from k to 2t/n

  int psi_print_sum_errors_k(char* wf,int k,ofstream& fr);
  //Print the sum of errors for errors from k to 2t/n in file fr
  
  int psi_print_u(char* wf, double ds,ofstream& fr, ofstream& fu);
  //Compute the u and print the sum of errors for errors from u to 2t/n
  //where u is such that P(N>u)<ds (and P(N>u-1)>ds)
  //ds = degree of significativity
  //File fr contains the results
  //File fu contains the u

  double psi_k_sum_errors
  (int k,double lx, double le, double p3, int n, double p, double par);
  //Compute and return the sum of errors for errors from 0 to k

  int psi_print_k_sum_errors(char* wf,int k,ofstream& fr);
  //Print the sum of errors for errors from 0 to k in file fr
   
  int psi_u_print(char* wf, double ds,ofstream& fr, ofstream& fu);
  //Compute the u and print the sum of errors for errors from 0 to u
  //where u is such that P(N<u)<ds (and P(N>u+1)>ds)
  //ds = degree of significativity
  //File fr contains the results
  //File fu contains the u 

  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // CHEN-STEIN Miguel  
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  
  //CHEN-STEIN MIGUEL

  double deltacs(double p, int n);
  //cf Thesis of Abadi
  //p is the probability of a word of size n

  double eps2cs(double p, int n);
  //cf Thesis of Abadi
  //p is the probability of a word of size n

  double chenstein_over_ma(char* wf, double ds, ofstream& fr, ofstream& fu);
  //Compute the Chen-Stein bound of Abadi and the u for overrepresented words  
  //ds = degree of significativity
  //File fr contains the results
  //File fu contains the u 

  double chenstein_under_ma(char* wf, double ds, ofstream& fr, ofstream& fu);
  //Compute the Chen-Stein bound of Abadi and the u for underrepresented words  
  //ds = degree of significativity
  //File fr contains the results
  //File fu contains the u 
 
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // Count of Words
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  int count_words(char* wf, char* sf, char * fa, ofstream& fu); 
  //Compute the number of words from wf in the sequence(s) sf 

};

//===========================================================================================
//===========================================================================================

#endif 
